import sys
from copy import deepcopy
from enum import Enum

from .proto_structures import EnumElement
from .proto_structures import ExtendElement
from .proto_structures import Extend
from .proto_structures import Import
from .proto_structures import Message
from .proto_structures import MessageElement
from .proto_structures import Option
from .proto_structures import Position
from .proto_structures import ProtoEnum
from .proto_structures import Service
from .proto_structures import ServiceElement
from .proto_structures import Oneof
from .util import to_lines


class Config:
    SPACES_BETWEEN_VALUE_COMMENT = 2
    SPACES_BEFORE_AFTER_EQUAL_SIGN = 1
    SPACES_BETWEEN_NUMBER_ANNOTATION = 1
    ONE_SPACE = ' '
    TOP_COMMENT_INDENTS = ' ' * 4

    def __init__(self,
                 indents_unit=2,
                 align_by_equal_sign=None,
                 top_comment=False,
                 flatten=False,
                 comment_max_length=None,
                 preview=False
                 ):
        self.indents_unit = indents_unit
        self.align_by_equal_sign = align_by_equal_sign
        self.top_comment = top_comment
        self.flatten = flatten
        self.comment_max_length = comment_max_length
        self.preview = preview


class Text:
    def __init__(self, text, color: int):
        self.text = text
        self.color = color

    def color_print(self):
        sys.stdout.write(u"\u001b[38;5;" + str(self.color) + "m" + self.text + u"\u001b[0m")


class ColorConfig(Enum):
    default = 15
    default_key_word = 11
    syntax_key = 11
    syntax_value = 10
    package_key = 11
    package_value = 15
    option_key = 11
    option_name = 15
    option_value = 40
    import_key = 11
    import_value = 40
    obj_key = 11
    obj_name = 15
    message_element_key = 196
    message_element_label = 11
    message_element_type = 150
    message_element_name = 140
    message_element_number = 12
    message_element_annotation = 15
    enum_element_name = 15
    enum_element_number = 12
    enum_element_annotation = 10
    service_element_lable = 11
    service_element_name = 140
    service_element_request = 15
    service_element_response = 15
    comment = 248


class Protobuf:

    def __init__(self):
        self.syntax = None
        self.package = None
        self.options = []
        self.imports = []
        self.objects = []
        self.text_list = []

        self.config = Config()

    def to_string(self,
                  indents=2,
                  align_by_equal_sign=None,
                  top_comment=False,
                  flatten=False,
                  comment_max_length=None,
                  preview=False
                  ):
        self.config = Config(
            indents_unit=indents,
            align_by_equal_sign=align_by_equal_sign,
            top_comment=top_comment,
            flatten=flatten,
            comment_max_length=comment_max_length,
            preview=preview
        )
        if self.config.flatten:
            self.flatten_objects()

        syntax_string = self.syntax_string()
        package_string = self.package_string()
        option_string = self.options_string()
        imports_string = self.imports_string()

        all_lines = []
        for obj in self.objects:
            lines = []
            self.format_object_without_comment(obj, lines, indents=0)
            max_length = self.get_max_length(lines)
            extra = self.get_max_lengths(lines)
            new_lines = []
            self.format_object(obj, new_lines, 0, extra)
            all_lines.append('\n'.join(new_lines))

        object_string = '\n\n'.join(all_lines)
        contents = [syntax_string, package_string, option_string, imports_string, object_string]
        contents = list(filter(None, contents))
        content = '\n\n'.join(contents)
        content = content + '\n'
        if self.config.preview:
            self.print_with_color()
        content_from_text_list = self.from_text_list()  # using formatted content build from text list.
        content_from_text_list = content_from_text_list.rstrip("\n")  # removed the two new lines("\n")
        content_from_text_list = content_from_text_list + "\n"  # append one new line("\n"), to keep the same as content

        return content  # content_from_text_list

    def print_with_color(self):
        print()
        for text in self.text_list:
            text.color_print()
        print()

    def from_text_list(self):
        content = ""
        for text in self.text_list:
            if isinstance(text, Text):
                debug = 0
            content = content + text.text
        return content

    def flatten_objects(self):
        new_objects = []

        for obj in self.objects:
            new_obj = deepcopy(obj)
            new_obj.elements = []
            for element in obj.elements:
                if isinstance(element, (Message, ProtoEnum, Service)):
                    self.flatten_object(element, new_objects)
                else:
                    new_obj.elements.append(element)

            new_objects.append(new_obj)

        self.objects = new_objects

    def flatten_object(self, obj, new_objects: list):
        new_obj = deepcopy(obj)
        new_obj.elements = []

        for element in obj.elements:
            if isinstance(element, (Message, ProtoEnum, Service)):
                self.flatten_object(element, new_objects)
            else:
                new_obj.elements.append(element)

        new_objects.append(new_obj)

    @staticmethod
    def get_max_length(lines):
        length = 0
        for line in lines:
            if length < len(line):
                length = len(line)

        return length

    def get_max_lengths(self, lines):
        max_equal_sign_index = 0
        max_length_of_number = 0
        max_length_of_object_line = 0
        max_length_of_service_element_line = 0
        s1 = ''
        s2 = ''
        for line in lines:
            if '=' in line:
                equa_sign_index = line.index('=')
                if max_equal_sign_index < equa_sign_index:
                    max_equal_sign_index = equa_sign_index
                    s1 = line[:equa_sign_index]

                semicolon_index = line.index(';')
                number_str = line[equa_sign_index + 1:semicolon_index].strip()
                number_length = len(number_str)
                if max_length_of_number < number_length:
                    max_length_of_number = number_length
                    s2 = number_str
            elif line.strip().startswith('rpc'):
                if max_length_of_service_element_line < len(line):
                    max_length_of_service_element_line = len(line)
            else:
                if max_length_of_object_line < len(line):
                    max_length_of_object_line = len(line)

        max_length_sample = f'{s1.rstrip()}{self.config.ONE_SPACE * self.config.SPACES_BEFORE_AFTER_EQUAL_SIGN}={self.config.ONE_SPACE * self.config.SPACES_BEFORE_AFTER_EQUAL_SIGN}{s2};'
        max_length = max(len(max_length_sample), max_length_of_service_element_line, max_length_of_object_line)

        return {
            'max_length': max_length,
            'max_equal_sign_index': max_equal_sign_index,
            'max_length_of_number': max_length_of_number,
            'max_length_of_object_line': max_length_of_object_line
        }

    def syntax_string(self):
        if not self.syntax:
            return ''

        line = f'{self.syntax.key} = "{self.syntax.value}";'
        color_text_list = [
            Text(self.syntax.key, ColorConfig.syntax_key.value),
            Text(" ", ColorConfig.default.value),
            Text("=", ColorConfig.default.value),
            Text(" ", ColorConfig.default.value),
            Text('"', ColorConfig.default.value),
            Text(self.syntax.value, ColorConfig.syntax_value.value),
            Text('"', ColorConfig.default.value),
            Text(";", ColorConfig.default.value),
        ]

        syntax_string = self.make_string(line, 0, self.syntax.comments, self.config.SPACES_BETWEEN_VALUE_COMMENT,
                                         color_text_list)
        self.text_list.append(Text("\n", ColorConfig.default.value))
        return syntax_string

    def package_string(self):
        if not self.package:
            return ''

        line = f'{self.package.key} {self.package.value};'
        color_text_list = [
            Text(self.package.key, ColorConfig.package_key.value),
            Text(" ", ColorConfig.default.value),
            Text(self.package.value, ColorConfig.syntax_value.value),
            Text(";", ColorConfig.default.value)
        ]

        package_string = self.make_string(line, 0, self.package.comments, self.config.SPACES_BETWEEN_VALUE_COMMENT,
                                          color_text_list)
        self.text_list.append(Text("\n", ColorConfig.default.value))
        return package_string

    def options_string(self):
        if not self.options:
            return ''

        max_length = self.max_length_of_option()

        string_list = []
        for option in self.options:
            string = self.option_string(option, max_length)
            string_list.append(string)

        self.text_list.append(Text("\n", ColorConfig.default.value))
        return '\n'.join(string_list)

    def max_length_of_option(self):
        max_length = 0
        for option in self.options:
            if option.value == 'true' or option.value == 'false':
                line = f'{option.key} {option.name} = {option.value};'
            else:
                line = f'{option.key} {option.name} = "{option.value}";'

            if max_length < len(line):
                max_length = len(line)

        return max_length

    def option_string(self, option: Option, max_length):
        if option.value == 'true' or option.value == 'false':
            line = f'{option.key} {option.name} = {option.value};'
            color_text_list = [
                Text(option.key, ColorConfig.option_key.value),
                Text(" ", ColorConfig.default.value),
                Text(option.name, ColorConfig.option_name.value),
                Text(" ", ColorConfig.default.value),
                Text("=", ColorConfig.default.value),
                Text(" ", ColorConfig.default.value),
                Text(option.value, ColorConfig.option_value.value),
                Text(";", ColorConfig.default.value)
            ]
        else:
            line = f'{option.key} {option.name} = "{option.value}";'
            color_text_list = [
                Text(option.key, ColorConfig.option_key.value),
                Text(" ", ColorConfig.default.value),
                Text(option.name, ColorConfig.option_name.value),
                Text(" ", ColorConfig.default.value),
                Text("=", ColorConfig.default.value),
                Text(" ", ColorConfig.default.value),
                Text('"', ColorConfig.default.value),
                Text(option.value, ColorConfig.option_value.value),
                Text('"', ColorConfig.default.value),
                Text(";", ColorConfig.default.value)
            ]

        space_between_number_comment = max_length - len(line) + self.config.SPACES_BETWEEN_VALUE_COMMENT

        return self.make_string(line, 0, option.comments, space_between_number_comment, color_text_list)

    def imports_string(self):
        if not self.imports:
            return ''

        max_length = self.max_length_of_import()

        string_list = []
        for obj in self.imports:
            string = self.import_string(obj, max_length)
            string_list.append(string)

        self.text_list.append(Text("\n", ColorConfig.default.value))
        return '\n'.join(string_list)

    def max_length_of_import(self):
        max_length = 0
        for obj in self.imports:
            line = f'{obj.key} "{obj.value}";'

            if max_length < len(line):
                max_length = len(line)

        return max_length

    def import_string(self, import_obj: Import, max_length):
        line = f'{import_obj.key} "{import_obj.value}";'
        color_text_list = [
            Text(import_obj.key, ColorConfig.import_key.value),
            Text(" ", ColorConfig.default.value),
            Text('"', ColorConfig.default.value),
            Text(import_obj.value, ColorConfig.import_value.value),
            Text('"', ColorConfig.default.value),
            Text(";", ColorConfig.default.value)
        ]
        space_between_number_comment = max_length - len(line) + self.config.SPACES_BETWEEN_VALUE_COMMENT

        return self.make_string(line, 0, import_obj.comments, space_between_number_comment, color_text_list)

    def create_object_header(self, obj, indents, no_comment, max_length):
        line = f'{obj.key} {obj.name} ' + "{"
        color_text_list = [
            Text(self.config.ONE_SPACE * indents, ColorConfig.default.value),
            Text(obj.key, ColorConfig.obj_key.value),
            Text(" ", ColorConfig.default.value),
            Text(obj.name, ColorConfig.obj_name.value),
            Text(" ", ColorConfig.default.value),
            Text("{", ColorConfig.default.value)
        ]

        space_between_number_comment = 2
        if max_length:
            space_between_number_comment = max_length - len(line) + self.config.SPACES_BETWEEN_VALUE_COMMENT

        if no_comment:
            return self.make_indented_line(line, indents=indents)
        else:
            comments = [comment for comment in obj.comments if comment.position is not Position.BOTTOM]
            return self.make_string(line, indents, comments, space_between_number_comment, color_text_list)

    def create_object_rear(self, obj, indents):
        bottom_comments = [comment for comment in obj.comments if comment.position == Position.BOTTOM]
        color_text_list = [
            Text(self.config.ONE_SPACE * indents, ColorConfig.default.value),
            Text("}", ColorConfig.default.value)
        ]

        rear = self.make_string('}', indents, bottom_comments, self.config.SPACES_BETWEEN_VALUE_COMMENT,
                                color_text_list)
        self.text_list.append(Text("\n", ColorConfig.default.value))
        return rear

    @staticmethod
    def get_object_element_class(obj):
        classes = {
            'message': MessageElement,
            'enum': EnumElement,
            'service': ServiceElement,
            'extend': MessageElement,
            'oneof': MessageElement
        }
        return classes[obj.key]

    def get_make_object_element_string_method(self, obj):
        methods = {
            Message.key: self.message_element_string,
            ProtoEnum.key: self.enum_element_string,
            Service.key: self.service_element_string,
            Extend.key: self.extend_element_string,
            Oneof.key: self.message_element_string
        }
        return methods[obj.key]

    def format_object(self, obj, string_list, indents, extra):
        max_length = extra['max_length']
        max_equal_sign_index = extra['max_equal_sign_index']
        max_length_of_number = extra['max_length_of_number']
        max_length_of_object_line = extra['max_length_of_object_line']

        message_header = self.create_object_header(obj, indents, False, max_length - indents)
        string_list.append(message_header)
        element_class = self.get_object_element_class(obj)
        element_str_method = self.get_make_object_element_string_method(obj)

        for element in obj.elements:
            if isinstance(element, element_class):
                line = element_str_method(
                    element,
                    indents + self.config.indents_unit,
                    True,
                    self.config.SPACES_BEFORE_AFTER_EQUAL_SIGN,
                    self.config.SPACES_BEFORE_AFTER_EQUAL_SIGN,
                    self.config.SPACES_BETWEEN_VALUE_COMMENT
                )

                if self.config.align_by_equal_sign:
                    line_without_number = line.split('=')[0].rstrip()
                    space_between_name_equal_sign = max_equal_sign_index - len(line_without_number)

                    if hasattr(element, 'number'):
                        space_between_number_comment = max_length - max_equal_sign_index - len('=') - len(';') - len(
                            element.number) - len(
                            element.annotation) - self.config.SPACES_BEFORE_AFTER_EQUAL_SIGN - self.config.SPACES_BETWEEN_NUMBER_ANNOTATION + self.config.SPACES_BETWEEN_VALUE_COMMENT

                        if element.annotation == '':
                            # the space between number and rules is tripped, so shouldn't count it
                            space_between_number_comment = space_between_number_comment + self.config.SPACES_BETWEEN_NUMBER_ANNOTATION

                    elif line.strip().startswith('rpc'):
                        space_between_number_comment = max_length - len(line) + self.config.SPACES_BETWEEN_VALUE_COMMENT
                    else:
                        space_between_number_comment = max_length - len(
                            line) - self.config.SPACES_BEFORE_AFTER_EQUAL_SIGN + self.config.SPACES_BETWEEN_VALUE_COMMENT

                    string = element_str_method(
                        element,
                        indents + self.config.indents_unit,
                        False,
                        space_between_name_equal_sign,
                        self.config.SPACES_BEFORE_AFTER_EQUAL_SIGN,
                        space_between_number_comment
                    )
                else:
                    space_between_number_comment = max_length - len(line) + self.config.SPACES_BETWEEN_VALUE_COMMENT
                    string = element_str_method(
                        element,
                        indents + self.config.indents_unit,
                        False,
                        self.config.SPACES_BEFORE_AFTER_EQUAL_SIGN,
                        self.config.SPACES_BEFORE_AFTER_EQUAL_SIGN,
                        space_between_number_comment
                    )

                string_list.append(string)
            else:
                self.format_object(element, string_list, indents + self.config.indents_unit, extra)

        message_rear = self.create_object_rear(obj, indents)
        string_list.append(message_rear)

    def format_object_without_comment(self, obj, string_list, indents):
        message_header = self.create_object_header(obj, no_comment=True, indents=indents, max_length=None)
        string_list.append(message_header)
        element_class = self.get_object_element_class(obj)
        element_str_method = self.get_make_object_element_string_method(obj)

        for element in obj.elements:
            if isinstance(element, element_class):
                string = element_str_method(
                    element,
                    indents + self.config.indents_unit,
                    True,
                    self.config.SPACES_BEFORE_AFTER_EQUAL_SIGN,
                    self.config.SPACES_BEFORE_AFTER_EQUAL_SIGN,
                    self.config.SPACES_BETWEEN_VALUE_COMMENT
                )
                string_list.append(string)
            else:
                self.format_object_without_comment(element, string_list, indents=indents + self.config.indents_unit)

        message_rear = self.make_indented_line('}', indents=indents)
        string_list.append(message_rear)

    def message_element_string(
            self,
            message_element: MessageElement,
            indents,
            no_comment,
            space_between_name_equal_sign,
            space_between_equal_sign_number,
            space_between_number_comment
    ):
        line = f'{message_element.label} {message_element.type} {message_element.name}{self.config.ONE_SPACE * space_between_name_equal_sign}={self.config.ONE_SPACE * space_between_equal_sign_number}{message_element.number}{self.config.ONE_SPACE * self.config.SPACES_BETWEEN_NUMBER_ANNOTATION}{message_element.annotation}'
        line = line.strip()  # remove prefix space when obj.label is empty string or rules is empty.
        line = f'{line};'
        color_text_list = [
            Text(self.config.ONE_SPACE * indents, ColorConfig.default.value),
            Text(message_element.label, ColorConfig.message_element_label.value),
            Text(" ", ColorConfig.default.value),
            Text(message_element.type, ColorConfig.message_element_type.value),
            Text(" ", ColorConfig.default.value),
            Text(message_element.name, ColorConfig.message_element_name.value),
            Text(self.config.ONE_SPACE * space_between_name_equal_sign, ColorConfig.default.value),
            Text("=", ColorConfig.default.value),
            Text(self.config.ONE_SPACE * space_between_equal_sign_number, ColorConfig.default.value),
            Text(message_element.number, ColorConfig.message_element_number.value),
            Text(self.config.ONE_SPACE * self.config.SPACES_BETWEEN_NUMBER_ANNOTATION, ColorConfig.default.value),
            Text(message_element.annotation, ColorConfig.message_element_annotation.value)
        ]
        if message_element.label.strip() == "":
            del color_text_list[1]
            del color_text_list[1]
        if message_element.annotation.strip() == "":
            del color_text_list[-1]
            del color_text_list[-1]
        color_text_list.append(Text(";", ColorConfig.default.value))

        if no_comment:
            return self.make_indented_line(line, indents)
        else:
            return self.make_string(line, indents, message_element.comments, space_between_number_comment,
                                    color_text_list)

    def extend_element_string(
            self,
            extend_element: ExtendElement,
            indents,
            no_comment,
            space_between_name_equal_sign,
            space_between_equal_sign_number,
            space_between_number_comment
    ):
        # using the message element making string
        return self.message_element_string(extend_element,
                                           indents,
                                           no_comment,
                                           space_between_name_equal_sign,
                                           space_between_equal_sign_number,
                                           space_between_number_comment)

    def enum_element_string(
            self,
            enum_element: EnumElement,
            indents,
            no_comment,
            space_between_name_equal_sign,
            space_between_equal_sign_number,
            space_between_number_comment
    ):
        line = f'{enum_element.name}{self.config.ONE_SPACE * space_between_name_equal_sign}={self.config.ONE_SPACE * space_between_equal_sign_number}{enum_element.number}{self.config.ONE_SPACE * self.config.SPACES_BETWEEN_NUMBER_ANNOTATION}{enum_element.annotation}'
        line = line.strip()  # remove prefix space when rules is empty.
        line = f'{line};'
        color_text_list = [
            Text(self.config.ONE_SPACE * indents, ColorConfig.default.value),
            Text(enum_element.name, ColorConfig.enum_element_name.value),
            Text(self.config.ONE_SPACE * space_between_name_equal_sign, ColorConfig.default.value),
            Text("=", ColorConfig.default.value),
            Text(self.config.ONE_SPACE * space_between_equal_sign_number, ColorConfig.default.value),
            Text(enum_element.number, ColorConfig.enum_element_number.value),
            Text(self.config.ONE_SPACE * self.config.SPACES_BETWEEN_NUMBER_ANNOTATION, ColorConfig.default.value),
            Text(enum_element.annotation, ColorConfig.enum_element_annotation.value),
        ]
        if enum_element.annotation.strip() == "":
            del color_text_list[-1]
            del color_text_list[-1]
        color_text_list.append(Text(";", ColorConfig.default.value))

        if no_comment:
            return self.make_indented_line(line, indents=indents)
        else:
            return self.make_string(line, indents, enum_element.comments, space_between_number_comment, color_text_list)

    def service_element_string(
            self,
            service_element: ServiceElement,
            indents,
            no_comment,
            space_between_name_equal_sign,
            space_between_equal_sign_number,
            space_between_number_comment
    ):
        # rpc SeatAvailability (SeatAvailabilityRequest) returns (SeatAvailabilityResponse);
        line = f'{service_element.label} {service_element.name} ({service_element.request}) returns ({service_element.response});'
        color_text_list = [
            Text(self.config.ONE_SPACE * indents, ColorConfig.default.value),
            Text(service_element.label, ColorConfig.service_element_lable.value),
            Text(" ", ColorConfig.default.value),
            Text(service_element.name, ColorConfig.service_element_name.value),
            Text(" ", ColorConfig.default.value),
            Text("(", ColorConfig.default.value),
            Text(service_element.request, ColorConfig.service_element_request.value),
            Text(")", ColorConfig.default.value),
            Text(" ", ColorConfig.default.value),
            Text("returns", ColorConfig.default_key_word.value),
            Text(" ", ColorConfig.default.value),
            Text("(", ColorConfig.default.value),
            Text(service_element.response, ColorConfig.service_element_response.value),
            Text(")", ColorConfig.default.value),
            Text(";", ColorConfig.default.value)
        ]

        if no_comment:
            return self.make_indented_line(line, indents=indents)
        else:
            return self.make_string(line, indents, service_element.comments, space_between_number_comment,
                                    color_text_list)

    def make_string(self, value_line, indents, comments, space_between_number_comment, color_text_list):
        if self.config.top_comment:
            formatted_comments = self.to_top_comment_lines(comments)
        else:
            formatted_comments = self.to_comment_lines(comments)
        top_comment_lines = formatted_comments["top"]
        bottom_comment_lines = formatted_comments["bottom"]
        right_comment = formatted_comments["right"]

        indented_top_comment_lines = []
        if top_comment_lines:
            comment_results = self.to_comment_lines_with_comment_slash(top_comment_lines, indents)
            indented_top_comment_lines = comment_results["indented_comment_lines"]
            comment_text_list = comment_results["comment_text_list"]
            color_text_list = comment_text_list + color_text_list

        indented_bottom_comment_lines = []
        if bottom_comment_lines:
            comment_results = self.to_comment_lines_with_comment_slash(bottom_comment_lines, indents)
            indented_bottom_comment_lines = comment_results["indented_comment_lines"]
            comment_text_list = comment_results["comment_text_list"]
            color_text_list = comment_text_list + color_text_list

        lines = indented_top_comment_lines
        if right_comment:
            line = f'{value_line}{self.config.ONE_SPACE * space_between_number_comment}// {right_comment}'
            lines.append(line)
            color_comment_text_list = [
                Text(self.config.ONE_SPACE * space_between_number_comment, ColorConfig.default.value),
                Text("//", ColorConfig.comment.value),
                Text(" ", ColorConfig.default.value),
                Text(right_comment, ColorConfig.comment.value),
                Text("\n", ColorConfig.default.value)
            ]
            color_text_list = color_text_list + color_comment_text_list
            self.text_list.extend(color_text_list)
        else:
            lines.append(value_line)
            # color_text_list.insert(0, Text(self.config.ONE_SPACE * indents, ColorConfig.default.value))
            color_text_list.append(Text("\n", ColorConfig.default.value))

            self.text_list.extend(color_text_list)

        if indented_bottom_comment_lines:
            lines.extend(indented_bottom_comment_lines)

        indented_lines = [f'{self.config.ONE_SPACE * indents}{line}' for line in lines]
        string = '\n'.join(indented_lines)

        return string

    def make_indented_line(self, line, indents=0):
        return f'{self.config.ONE_SPACE * indents}{line}'

    def convert_to_lines(self, comment):
        """
        convert comment text to multiple lines if the comment_max_length flag is set and the length of the text exceeds
        it.
        :param comment: comment object.
        :return: string lines of comment text.
        """
        text_lines = [line.strip() for line in comment.text.split('\n')]

        new_text_lines = []
        if self.config.comment_max_length is not None:
            for line in text_lines:
                new_text_lines.extend(to_lines(line, self.config.comment_max_length))
            text_lines = new_text_lines
        return text_lines

    def to_top_comment_lines(self, comments):
        """
        convert right comment to top comment lines:
        (1) keep top comments. If the comment_max_length flag is set and the length of one original line exceeds it,
            this line will be converted to multiple lines.
        (2) convert right comment to multiple lines, append it to top comments lines.
        (3) convert bottom comments(the comments under the last element or follow the message etc.) to multiple lines.
            If the comment_max_length flag is set and the length of one original line exceeds it, this line will be
            converted to multiple lines. This lines will not be appended to top comments lines.
        :param comments: all comments of an object.
        :return: formatted comments lines.
        """
        top_comment_lines = []
        right_comment = ''
        bottom_comment_lines = []
        for comment in comments:
            if comment.position == Position.TOP:
                top_comment_lines.extend(self.convert_to_lines(comment))
            if comment.position == Position.BOTTOM:
                bottom_comment_lines.extend(self.convert_to_lines(comment))
            if comment.position == Position.Right:
                line = comment.text
                text_lines = [line]

                new_text_lines = []
                if self.config.comment_max_length is not None:
                    for l in text_lines:
                        new_text_lines.extend(to_lines(l, self.config.comment_max_length))

                if self.config.comment_max_length is not None:
                    top_comment_lines.extend(new_text_lines)
                else:
                    top_comment_lines.extend(text_lines)

                right_comment = ''
        results = {
            "top": top_comment_lines,
            "right": right_comment,
            "bottom": bottom_comment_lines
        }
        return results

    def to_comment_lines(self, comments):
        """
        convert right comment to comment lines:
        (1) keep top comments. If the comment_max_length flag is set and the length of one original line exceeds it,
            this line will be converted to multiple lines.
        (2) convert right comment to multiple lines.
        (3) convert bottom comments(the comments under the last element or follow the message etc.) to multiple lines.
            If the comment_max_length flag is set and the length of one original line exceeds it, this line will be
            converted to multiple lines. This lines will not be appended to top comments lines.
        :param comments: all comments of an object.
        :return: formatted comments lines.
        """
        top_comment_lines = []
        right_comment = ''
        bottom_comment_lines = []
        for comment in comments:
            if comment.position == Position.TOP:
                top_comment_lines.extend(self.convert_to_lines(comment))
            if comment.position == Position.BOTTOM:
                bottom_comment_lines.extend(self.convert_to_lines(comment))
            if comment.position == Position.Right:
                right_comment = comment.text
        results = {
            "top": top_comment_lines,
            "right": right_comment,
            "bottom": bottom_comment_lines
        }
        return results

    def to_comment_lines_with_comment_slash(self, comment_lines, indents):
        indented_comment_lines = []
        comment_text_list = []
        for comment_line in comment_lines:
            indented_comment_lines.append(f'**{self.config.TOP_COMMENT_INDENTS}{comment_line}')
            one_comment_text_list = [
                Text(self.config.ONE_SPACE * indents, ColorConfig.default.value),
                Text("**", ColorConfig.comment.value),
                Text(self.config.TOP_COMMENT_INDENTS, ColorConfig.default.value),
                Text(comment_line, ColorConfig.comment.value),
                Text("\n", ColorConfig.default.value)
            ]
            comment_text_list = comment_text_list + one_comment_text_list
        indented_comment_lines.insert(0, '/*')
        indented_comment_lines.append('*/')

        comment_header_text_list = [
            Text(self.config.ONE_SPACE * indents, ColorConfig.default.value),
            Text("/*", ColorConfig.comment.value),
            Text("\n", ColorConfig.default.value)
        ]
        comment_rear_text_list = [
            Text(self.config.ONE_SPACE * indents, ColorConfig.default.value),
            Text("*/", ColorConfig.comment.value),
            Text("\n", ColorConfig.default.value)
        ]
        comment_text_list = comment_header_text_list + comment_text_list + comment_rear_text_list
        result = {
            "indented_comment_lines": indented_comment_lines,
            "comment_text_list": comment_text_list
        }
        return result



SLASH_STAR = '/*'
STAR_SLASH = '*/'
STAR = '*'
DOUBLE_SLASH = '//'
LEFT_BRACE = '{'
RIGHT_BRACE = '}'
LEFT_SQUARE_BRACKET = '['
RIGHT_SQUARE_BRACKET = ']'
EQUAL_SIGN = '='
SEMICOLON = ';'
LEFT_ANGLE_BRACKET = '<'
RIGHT_ANGLE_BRACKET = '>'



from .comment import CommentParser
from .constant import SEMICOLON
from .proto_structures import Import
from .protobuf import Protobuf
from .util import strip


class ImportParser(CommentParser):

    @classmethod
    def parse_and_add(cls, proto_obj: Protobuf, line, top_comments):
        new_import = cls._parse(line, top_comments)
        cls._check(new_import, proto_obj)
        proto_obj.imports.append(new_import)

    @classmethod
    def _check(cls, new_import, proto_obj):
        for existing_import in proto_obj.imports:
            if existing_import.value == new_import.value:
                raise Exception(f'duplicate import detected: {new_import.value}!')

    @classmethod
    def _parse(cls, line, top_comments):
        value = cls._get_value(line)
        comments = cls.create_comment(line, top_comments)
        n_import = Import(value, comments)
        return n_import

    @classmethod
    def _get_value(cls, line):
        line = line.strip()
        value = strip(line[len('import '):line.index(SEMICOLON)])
        return value



from enum import Enum


class Position(Enum):
    LEFT = 'left'
    Right = 'right'
    TOP = 'top'
    BOTTOM = 'bottom'


class Comment:
    def __init__(self, text: str, position: Position):
        self.text = text
        self.position = position


class Syntax:
    key = "syntax"

    def __init__(self, value, comments: list):
        self.value = value
        self.comments = comments


class Package:
    key = "package"

    def __init__(self, value, comments: list):
        self.value = value
        self.comments = comments


class Import:
    key = "import"

    def __init__(self, value, comments: list):
        self.value = value
        self.comments = comments


class Option:
    key = "option"

    def __init__(self, name, value, comments: list):
        self.name = name
        self.value = value
        self.comments = comments


class Element:
    def __init__(self, type, name, number, annotation='', label='', comments=[]):
        self.label = label
        self.type = type
        self.name = name
        self.number = number
        self.annotation = annotation
        self.comments = comments


class MessageElement:
    def __init__(self, type, name, number, annotation='', label='', comments=[]):
        self.label = label
        self.type = type
        self.name = name
        self.number = number
        self.annotation = annotation
        self.comments = comments


class Message:
    key = "message"

    def __init__(self, name, elements=None, comments=None):
        if elements is None:
            elements = []
        if comments is None:
            comments = []
        self.name = name
        self.elements = elements
        self.comments = comments


class ExtendElement:
    def __init__(self, type, name, number, annotation='', label='', comments=None):
        if comments is None:
            comments = []
        self.label = label
        self.type = type
        self.name = name
        self.number = number
        self.annotation = annotation
        self.comments = comments


class Extend:
    key = "extend"

    def __init__(self, name, elements=None, comments=None):
        if comments is None:
            comments = []
        if elements is None:
            elements = []

        self.name = name
        self.elements = elements
        self.comments = comments


class OneofElement:
    def __init__(self, type, name, number, annotation='', label='', comments=None):
        if comments is None:
            comments = []
        self.label = label
        self.type = type
        self.name = name
        self.number = number
        self.annotation = annotation
        self.comments = comments


class Oneof:
    key = "oneof"

    def __init__(self, name, elements=None, comments=None):
        if comments is None:
            comments = []
        if elements is None:
            elements = []

        self.name = name
        self.elements = elements
        self.comments = comments


class EnumElement:
    def __init__(self, name, number, annotation='', comments=None):
        if comments is None:
            comments = []
        self.name = name
        self.number = number
        self.annotation = annotation
        self.comments = comments


class ProtoEnum:
    key = "enum"

    def __init__(self, name, elements=None, comments=None):
        if comments is None:
            comments = []
        if elements is None:
            elements = []

        self.name = name
        self.elements = elements
        self.comments = comments


class ServiceElement:
    def __init__(self, label, name, request, response, comments=None):
        if comments is None:
            comments = []
        self.label = label
        self.name = name
        self.request = request
        self.response = response
        self.comments = comments


class Service:
    key = "service"

    def __init__(self, name, elements=None, comments=None):
        if comments is None:
            comments = []
        if elements is None:
            elements = []
        self.name = name
        self.elements = elements
        self.comments = comments



from .comment import CommentParser
from .constant import (
    DOUBLE_SLASH,
    SLASH_STAR,
    LEFT_BRACE,
    RIGHT_BRACE,
    EQUAL_SIGN,
    SEMICOLON
)


class ProtoTypeDetector:

    @staticmethod
    def get_top_comments(lines):
        comment_parser = CommentParser()
        comments_lines = comment_parser.pick_up_comment(lines)
        comments = comment_parser.parse(comments_lines)

        return comments

    def get_type(self, line):
        if self.is_syntax_line(line):
            return 'syntax'
        if self.is_package_line(line):
            return 'package'
        if self.is_option_line(line):
            return 'option'
        if self.is_import_line(line):
            return 'import'
        if self.is_message_object(line):
            return 'message'
        if self.is_enum_object(line):
            return 'enum'
        if self.is_service_object(line):
            return 'service'
        if self.is_extend_object(line):
            return 'extend'
        if self.is_message_element(line):
            return 'message_element'
        # if self.is_extend_element(line):
        #     return 'extend_element'
        if self.is_enum_element(line):
            return 'enum_element'
        if self.is_service_element(line):
            return 'service_element'
        if self.is_oneof_object(line):
            return 'oneof'

        return 'unknown'

    @staticmethod
    def is_object_type(proto_type):
        return proto_type in ['message', 'enum', 'service', 'oneof']

    @staticmethod
    def is_syntax_line(line):
        return line.replace(' ', '').startswith('syntax=')

    @staticmethod
    def is_package_line(line):
        return line.strip().startswith('package ')

    @staticmethod
    def is_option_line(line):
        return line.strip().startswith('option ')

    @staticmethod
    def is_import_line(line):
        return line.strip().startswith('import ')

    @staticmethod
    def check_object(line, flag):
        if line.count(flag) == 0:
            return False

        if line.count(DOUBLE_SLASH) > 0 and line.index(flag) > line.index(DOUBLE_SLASH):
            return False

        if line.count(SLASH_STAR) > 0 and line.index(flag) > line.index(SLASH_STAR):
            return False

        return True

    def is_object_start(self, line):
        return self.check_object(line, LEFT_BRACE)

    def is_object_end(self, line):
        return self.check_object(line, RIGHT_BRACE)

    @staticmethod
    def is_message_object(line):
        return line.strip().startswith('message ') and line.strip().count(LEFT_BRACE)

    @staticmethod
    def is_extend_object(line):
        return line.strip().startswith('extend ') and line.strip().count(LEFT_BRACE)

    @staticmethod
    def is_oneof_object(line):
        return line.strip().startswith('oneof ') and line.strip().count(LEFT_BRACE)

    def is_message_element(self, line):
        if not self.is_element_line(line):
            return False

        if self.is_map_element(line):
            return True

        parts = line.strip().split(EQUAL_SIGN)
        parts = [e for e in parts[0].strip().split(' ')]
        parts = list(filter(None, parts))

        if parts[0] == 'rpc':  # service element
            return False

        return len(parts) >= 2

    def is_element_line(self, line):
        if line.count(SEMICOLON) == 0:
            return False

        if line.count(DOUBLE_SLASH) > 0 and line.index(SEMICOLON) > line.index(DOUBLE_SLASH):
            return False

        if line.count(SLASH_STAR) > 0 and line.index(SEMICOLON) > line.index(SLASH_STAR):
            return False

        if self.is_service_element(line):
            return True

        if self.is_map_element(line):
            return True

        return line.strip().count(SEMICOLON) > 0 and line.strip().count(EQUAL_SIGN) > 0

    @staticmethod
    def is_service_element(line):
        # the service element looks like:
        # rpc SeatAvailability (SeatAvailabilityRequest) returns (SeatAvailabilityResponse);
        line = line.strip()
        return line.startswith('rpc ')

    @staticmethod
    def is_map_element(line):
        # the map element looks like:
        # map<string, Project> projects = 3;
        line = line.strip().replace(' ', '')
        return line.startswith('map<')

    @staticmethod
    def is_enum_object(line):
        return line.strip().startswith('enum ') and line.strip().count(LEFT_BRACE)

    def is_enum_element(self, line):
        if not self.is_element_line(line):
            return False

        parts = line.strip().split(EQUAL_SIGN)
        parts = [e for e in parts[0].strip().split(' ')]
        parts = list(filter(None, parts))

        return len(parts) == 1

    @staticmethod
    def is_service_object(line):
        return line.strip().startswith('service ') and line.strip().count(LEFT_BRACE)



from .comment import CommentParser
from .type_detector import ProtoTypeDetector
from .import_parser import ImportParser
from .object_parser import ObjectParser
from .option_parser import OptionParser
from .package_parser import PackageParser
from .protobuf import Protobuf
from .syntax_parser import SyntaxParser


class ProtoParser(CommentParser):
    PARSERS = {
        'syntax': SyntaxParser,
        'package': PackageParser,
        'option': OptionParser,
        'import': ImportParser,
        'message': ObjectParser,
        'enum': ObjectParser,
        'extend': ObjectParser,
        'service': ObjectParser,
        'oneof': SyntaxParser
    }

    @staticmethod
    def load(fp):
        """
        Load proto from file, deserialize it as a Protobuf object.
        :param fp: The absolute file path of the proto file.
        :return: Object of Protobuf.
        """
        lines = ProtoParser._read_lines(fp)
        return ProtoParser._parse(lines)

    @staticmethod
    def loads(proto_str):
        """
        Parse proto string, return a Protobuf object.
        :param proto_str: The proto string need to parse.
        :return: Object of Protobuf.
        """
        lines = proto_str.split('\n')
        return ProtoParser._parse(lines)

    @staticmethod
    def _read_lines(path):
        """
        Read data from file as line list, all blank and empty lines are removed before and after valid text.
        :param path: Absolute file path.
        :return: Line list of the striped file content.
        """
        with open(path) as f:
            content = f.read()
            content = content.strip()
            lines = content.split('\n')
            return lines

    @staticmethod
    def _parse(lines, protobuf_obj=None):
        """
        Parse proto content lines, deserialize them as a Protobuf object.
        :param lines: the proto content lines.
        :return: an object of Protobuf.
        """
        if protobuf_obj is None:
            protobuf_obj = Protobuf()

        top_comments = ProtoTypeDetector().get_top_comments(lines)
        if len(lines) == 0:
            return protobuf_obj

        first_line = lines[0]
        proto_type = ProtoTypeDetector().get_type(first_line)

        parser = ProtoParser.PARSERS[proto_type]()
        if parser is None:
            return protobuf_obj

        if isinstance(parser, ObjectParser):
            parser.parse(lines)
            top_comments = ProtoParser.create_comment(first_line, top_comments)
            all_comments = top_comments + parser.objects[0].comments
            parser.objects[0].comments = all_comments
            protobuf_obj.objects.extend(parser.objects)
        else:
            line = lines.pop(0)
            parser.parse_and_add(protobuf_obj, line, top_comments)

        if len(lines) == 0:
            return protobuf_obj

        return ProtoParser._parse(lines, protobuf_obj)



from .comment import CommentParser
from .protobuf import Protobuf
from .proto_structures import Syntax
from .util import strip


class SyntaxParser(CommentParser):

    @classmethod
    def parse_and_add(cls, proto_obj: Protobuf, line, top_comment_list):
        cls._check(proto_obj)

        proto_obj.syntax = cls._parse(line, top_comment_list)

    @classmethod
    def _check(cls, proto_obj):
        if proto_obj.syntax is not None:
            raise Exception('duplicate package detected!')

    @classmethod
    def _parse(cls, line, top_comment_list):
        value = cls._get_value(line)
        comments = cls.create_comment(line, top_comment_list)
        syntax = Syntax(value, comments)
        return syntax

    @classmethod
    def _get_value(cls, line):
        line = line.strip().replace(' ', '')
        value = strip(line[len('syntax='):line.index(';')])
        return value



import curses
import sys

from docutils.parsers.rst.directives.tables import align


def remove_prefix(text, prefix):
    if text.startswith(prefix):
        return text[len(prefix):]
    return text


def remove_suffix(text, suffix):
    if text.endswith(suffix):
        return text[:-len(suffix)]
    return text


def is_word(line, target_index):
    return line[target_index + 1].lower() not in ' \n' or line[target_index - 1].lower() not in ' \n'


def find_word_start_index(line, target_index):
    index = target_index - 1
    while index >= 0:
        if line[index].lower() in ' \n':
            return index
        index = index - 1
    return target_index


def find_word_end_index(line, target_index):
    index = target_index + 1
    while index <= len(line):
        if line[index].lower() in ' \n':
            return index
        index = index + 1
    return target_index


def to_lines(line, length):
    lines = []
    while True:
        if len(line) <= length:
            lines.append(line)
            break

        if not is_word(line, length - 1):
            lines.append(line[:length - 1])
            sub_line = line[:length - 1]
            line = line[length - 1:].strip()
        else:
            index = find_word_start_index(line, length - 1)
            sub_line = line[:index]
            line = line[index:].strip()

        lines.append(sub_line)

    return lines


def read_file(file_path):
    with open(file_path) as f:
        return f.read()


def view_curses_colors():
    def print_color(stdscr):
        curses.start_color()
        curses.use_default_colors()
        for i in range(0, curses.COLORS):
            curses.init_pair(i + 1, i, -1)
        try:
            for i in range(0, 255):
                stdscr.addstr(str(i) + " ", curses.color_pair(i))
        except curses.ERR:
            # End of screen reached
            pass
        stdscr.getch()

    curses.wrapper(print_color)


def color_print(msg):
    def func(stdscr):
        stdscr.scrollok(True)  # enable scrolling, so it can print string with new lines

        curses.start_color()
        curses.use_default_colors()

        curses.init_pair(124, 123, -1)  # changes the definition of color pair 124
        curses.init_pair(197, 196, -1)  # changes the definition of color pair 197
        try:
            stdscr.addstr(msg + " ", curses.color_pair(124))
            stdscr.addstr("\n\npress any key to exit view", curses.color_pair(197))  # red text
        except curses.ERR:
            # End of screen reached
            pass
        stdscr.getch()

    curses.wrapper(func)


def proto_print(msg):
    sys.stdout.write(u"\u001b[38;5;" + str(229) + "m " + msg)
    print(u"\u001b[0m")


def print_info(msg):
    sys.stdout.write(u"\u001b[38;5;" + str(38) + "m " + msg)
    print(u"\u001b[0m")


def show_colors():
    for i in range(0, 16):
        for j in range(0, 16):
            code = str(i * 16 + j)
            sys.stdout.write(u"\u001b[38;5;" + code + "m " + code.ljust(4))
    print(u"\u001b[0m")  # clear


def split(line: str, x: int):
    """
    split a line every x characters
    :param line: original line need to be split.
    :param x: max length of split lines
    :return: a string list that all elements are not more than x characters.
    """
    return [line[i: i + x] for i in range(0, len(line), x)]


def add_prefix(lines: [str], prefix: str):
    return [prefix + line for line in lines]


def add_suffix(lines: [str], suffix: str):
    return [line + suffix for line in lines]


def align_lines_with_filler(text: str, length: int, filler: str = " ", align='left'):
    """
    align text
    :param text: text to be formatted.
    :param length: max length limitation for each line.
    :param filler: the character used to added to the line to make sure the length is the same for lines.
    :param align: the way of alignment, left, right or center
    :return: formatted string lines
    """
    lines = text.split("\n")

    formatted_lines = []
    for line in lines:
        formatted_lines.extend(to_lines(line, length))

    lines = formatted_lines
    if align == 'left':
        return [f"{line.strip()}{filler * (length - len(line.strip()))}" for line in lines]

    if align == 'right':
        return [f"{filler * (length - len(line.strip()))}{line.strip()}" for line in lines]

    if align == 'center':
        processed_lines = []
        for line in lines:
            filler_amount = int((length - len(line.strip())) / 2)
            if (length - len(line.strip())) % 2 == 1:
                line = f"{filler * filler_amount}{line.strip() + ' '}{filler * filler_amount}"
            else:
                line = f"{filler * filler_amount}{line.strip()}{filler * filler_amount}"
            processed_lines.append(line)
        return processed_lines

    return lines


def replace_at_start(line: str, new: str):
    """
    replace the start substring with same length of new string from the start of the line with the new string.
    :param line: lien to be processed.
    :param new: new string.
    :return: replaced line.
    """
    if len(line) <= len(new):
        return new[:len(line)]
    else:
        return f"{new}{line[len(new):]}"


def strip(s: str):
    """
    Return a copy of the string with leading and trailing whitespace, double quotation marks and single quotation marks
    removed.
    :param s: string to be stripped.
    :return: new copy of tripped string.
    """
    return s.strip().replace('"', "").replace("'", "")



from wrap_enum import WrapEnum


# https://developers.google.com/protocol-buffers/docs/proto3
class BasicValueType(WrapEnum):
    DOUBLE = 'double'
    FLOAT = 'float'
    INT32 = 'int32'
    INT64 = 'int64'
    UINT32 = 'uint32'
    UINT64 = 'uint64'
    SINT32 = 'sint32'
    SINT64 = 'sint64'
    FIXED32 = 'fixed32'
    FIXED64 = 'fixed64'
    SFIXED32 = 'sfixed32'
    SFIXED64 = 'sfixed64'
    BOOL = 'bool'
    STRING = 'string'
    BYTES = 'bytes'



from .comment import CommentParser
from .constant import EQUAL_SIGN, SEMICOLON
from .proto_structures import Option
from .protobuf import Protobuf
from .util import strip


class OptionParser(CommentParser):

    @classmethod
    def parse_and_add(cls, proto_obj: Protobuf, line, top_comments):
        new_option = cls._parse(line, top_comments)
        cls._check(new_option, proto_obj)
        proto_obj.options.append(new_option)

    @classmethod
    def _check(cls, new_option, proto_obj: Protobuf):
        for option in proto_obj.options:
            if option.name == new_option.name:
                raise Exception(f'duplicate option detected: {new_option.name}!')

    @classmethod
    def _parse(cls, line, top_comments):
        name, value = cls._get_value(line)
        comments = cls.create_comment(line, top_comments)
        option = Option(name, value, comments)
        return option

    @classmethod
    def _get_value(cls, line):
        line = line.strip()
        lindex = len('option ')
        equal_sign_index = line.index(EQUAL_SIGN)
        semicolon_index = line.index(SEMICOLON)
        name = strip(line[lindex:equal_sign_index])
        value = strip(line[equal_sign_index + 1:semicolon_index])

        return name, value



import setuptools

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

install_requires = [
    'attrdict>=2.0.1',
    'docutils'
]

dev_requires = [
    'build==0.5.1',
    'twine>=3.4.2,<4',
    'pip-tools>=5,<6',
    'pytest>=6.2.4,<6.3',
    'attrdict>=2.0.1',
]

setuptools.setup(
    name="proto-formatter",
    version="0.2.1",
    author="YiXiaoCuoHuaiFenZi",
    author_email="249664317@qq.com",
    description="Protocol Buffers file formatter.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/YiXiaoCuoHuaiFenZi/proto-formatter",
    project_urls={
        "Bug Tracker": "https://github.com/YiXiaoCuoHuaiFenZi/proto-formatter/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
    install_requires=install_requires,
    setup_requires=[],
    tests_require=[],
    extras_require={
        "dev": dev_requires
    },
    entry_points={
        'console_scripts': ['proto_formatter=proto_formatter.main:main'],
    },
)



from .comment import CommentParser
from .proto_structures import Package
from .protobuf import Protobuf


class PackageParser(CommentParser):

    @classmethod
    def parse_and_add(cls, proto_obj: Protobuf, line, top_comments):
        cls._check(proto_obj)
        proto_obj.package = cls._parse(line, top_comments)

    @classmethod
    def _check(cls, proto_obj):
        if proto_obj.package is not None:
            raise Exception('duplicate package detected!')

    @classmethod
    def _parse(cls, line, top_comments):
        value = cls._get_value(line)
        comments = CommentParser.create_comment(line, top_comments)
        package = Package(value, comments)
        return package

    @classmethod
    def _get_value(cls, line):
        line = line.strip()
        value = line[len('package '):line.index(';')].strip()
        return value
